/* ====================================================================
 * Copyright (c) 1996 Vidya Media Ventures, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of this source code or a derived source code must
 *    retain the above copyright notice, this list of conditions and the
 *    following disclaimer. 
 *
 * 2. Redistributions of this module or a derived module in binary form
 *    must reproduce the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY VIDYA MEDIA VENTURES, INC. ``AS IS'' AND 
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL VIDYA MEDIA VENTURES, INC.
 * OR ITS EMPLOYEES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 * $Id: mod_auth_pg.c,v 1.2 2004/10/23 01:47:17 babolo Exp $
 *
 * This software is a contribution to and makes use of the Apache HTTP
 * server which is written, maintained and copywritten by the Apache Group.
 * See http://www.apache.org/ for more information.
 *
 * This software makes use of libpq which an interface to the PostgreSQL
 * database.  PostgreSQL is copyright (c) 1994 by the Regents of the
 * University of California.  As of this writing, more information on
 * PostgreSQL can be found at http://www.postgresSQL.org/
 *
 */

/*
 * 
 * PostgreSQL authentication
 *
 *
 * Needs libpq-fe.h and libpq.a
 *
 * Outline:
 *
 * - Authentication
 *   One database, and one (or two) tables.  One table holds the username and
 *   the encryped (or plain) password.  The other table holds the username and the names
 *   of the group to which the user belongs.  It is possible to have username,
 *   groupname and password in the same table.
 * - Access Logging
 *   Every authentication access is logged in the same database of the 
 *   authentication table, but in different table.
 *   User name and date of the request are logged.
 *   As option, it can log password, ip address, request line.
 *
 * Module Directives:  See html documentation
 * 
 * Changelog: See html documentation
 *
 * see http://www.postgreSQL.org/
 *
 *
 *
 *		Homepage	http://www.giuseppetanzilli.it/mod_auth_pg/
 *
 *		Latest sources  http://www.giuseppetanzilli.it/mod_auth_pg/dist/
 *
 *		Maintainer:
 *		@BABOLO <.@babolo.ru>		
 *
 *		Original source: 
 * 		Adam Sussman (asussman@vidya.com) Feb, 1996
 *		Matthias Eckermann
 *		eckerman@lrz.uni-muenchen.de
 *		Giuseppe Tanzilli (g.tanzilli@gruppocsf.com)
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <httpd.h>
#include <http_config.h>
#include <http_core.h>
#include <http_log.h>
#include <http_protocol.h>
#include <ap_md5.h>
#include <ap_alloc.h>
#include <libpq-fe.h>

#define	AUTH_PGSQL_VERSION VERS

/* We need Apache 1.3.1 */
#if (MODULE_MAGIC_NUMBER >= 19980713)
# define _const const
# define IS_APACHE_13 1
#endif

#ifndef MAX_STRING_LEN
# define MAX_STRING_LEN 3000
#endif

enum classes {sq, dq, bs, sy, nl, xx};
enum states {c0, c1, cn, q0, q1, qn, k0, k1, k2, kn, s0, s1, s2, sn, ex};
enum contx {SINGLE_QU, DOUBLE_QU};

typedef struct {
    char *auth_pg_host;
    char *auth_pg_database;
    char *auth_pg_port;
    char *auth_pg_options;
    char *auth_pg_user;
    char *auth_pg_pwd;
    char *auth_pg_function;
    char *auth_pg_context;
    char *auth_pg_finish;
    int auth_pg_authoritative;
} pg_auth_config_rec;

static pool *auth_pg_pool = NULL;
module MODULE_VAR_EXPORT auth_pg_module;
char pg_errstr[MAX_STRING_LEN];
/* global errno to be able to handle config/sql 
 * failures separately
 */
PGconn *pg_conn = NULL;

#ifdef DEBUG
# include <time.h>
# define DL(A) {FILE*DBG = fopen("/var/log/dbg","a");fprintf(DBG,"%s %5d %s\n",pt(),getpid(),A);fclose(DBG);}

char *
pt() {
    time_t t;
    char *c;
    t = time(NULL);
    c = ctime(&t);
    c[24] = '\0';
    return(c);
}
#else
# define DL(A)
#endif

size_t
wordscan(enum contx q, const char *from, char **to, const char *p) {
    size_t n;
    const char *in;
    enum states state;
    u_int32_t control = 0;
char *t = to ? *to : NULL;
int i;
    const enum classes class[256] =
    {nl,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,dq,sy ,sy,sy,sy,sq ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy

    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,bs,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy

    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy

    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy ,sy,sy,sy,sy
    };
#define C_a2 0x40000000 /* ++2                 */
#define C_a1 0x20000000 /* ++1                 */
#define C_bs 0x10000000 /* \\                  */
#define C_sq  0x8000000 /* copy '              */
#define C_dq  0x4000000 /* copy "              */
#define C_sk  0x2000000 /* copy '              */
#define C_dk  0x1000000 /* copy "              */
#define C_cp   0x800000 /* copy symbol         */
#define C_wp   0x400000 /* switch to p         */
#define C_u0   0x200000 /* in++                */
#define C_state    0xFF /*   */
#define C_bc C_bs|C_cp
    const u_int32_t automa[ex][xx] =
    /*               sq,                dq,                bs,           sy,                nl */
    {{     C_a2|C_u0|c0,      C_a1|C_u0|c0,      C_a2|C_u0|c0, C_a1|C_u0|c0,      C_a1|C_wp|c1}/*c0*/
    ,{     C_a1|C_u0|c1,      C_a1|C_u0|c1,      C_a1|C_u0|c1, C_a1|C_u0|c1,                ex}/*c1*/
    ,{     C_a1|C_u0|cn,      C_a1|C_u0|cn,      C_a1|C_u0|cn, C_a1|C_u0|cn,           C_wp|c1}/*cn*/
    ,{     C_a1|C_u0|q0,      C_a2|C_u0|q0,      C_a2|C_u0|q0, C_a1|C_u0|q0,      C_a1|C_wp|q1}/*q0*/
    ,{     C_a1|C_u0|q1,      C_a1|C_u0|q1,      C_a1|C_u0|q1, C_a1|C_u0|q1,                ex}/*q1*/
    ,{     C_a1|C_u0|qn,      C_a1|C_u0|qn,      C_a1|C_u0|qn, C_a1|C_u0|qn,           C_wp|q1}/*qn*/
    ,{C_sq|C_bs|C_cp|k1, C_sq     |C_cp|k1, C_sq|C_bs|C_cp|k1, C_sq|C_cp|k1, C_sq|C_sk|C_wp|k2}/*k0*/
    ,{     C_bs|C_cp|k1,           C_cp|k1,      C_bs|C_cp|k1,      C_cp|k1,      C_sk|C_wp|k2}/*k1*/
    ,{          C_cp|k2,           C_cp|k2,           C_cp|k2,      C_cp|k2,                ex}/*k2*/
    ,{          C_cp|kn,           C_cp|kn,           C_cp|kn,      C_cp|kn,           C_wp|k2}/*kn*/
    ,{C_dq     |C_cp|s1, C_dq|C_bs|C_cp|s1, C_dq|C_bs|C_cp|s1, C_dq|C_cp|s1, C_dq|C_dk|C_wp|s2}/*s0*/
    ,{          C_cp|s1,      C_bs|C_cp|s1,      C_bs|C_cp|s1,      C_cp|s1,      C_dk|C_wp|s2}/*s1*/
    ,{          C_cp|s2,           C_cp|s2,           C_cp|s2,      C_cp|s2,                ex}/*s2*/
    ,{          C_cp|sn,           C_cp|sn,           C_cp|sn,      C_cp|sn,           C_wp|s2}/*sn*/
    };
    if  (q == SINGLE_QU) {
        if  (from == NULL) {
            from = "NULL";
            state = (to) ? kn : cn;
            n = 0;
        } else {
            state = (to) ? k0 : c0;
            n = 1;
        }
    } else if (q == DOUBLE_QU) {
        if  (from == NULL) {
            from = "NULL";
            state = (to) ? sn : qn;
            n = 0;
        } else {
            state = (to) ? s0 : q0;
            n = 1;
        }
    } else {                                                /* Low priority to display - impossible */
        snprintf(pg_errstr, MAX_STRING_LEN, "auth_pg: Internal error");
        return(0);
    };
    for (in = from; state < ex;) {
        control = automa[state][class[*in]];
        if  (control & C_a1) n += 1;
        if  (control & C_a2) n += 2;
        if  (control & C_bs) *(*to)++ = '\\';
        if  (control & C_sq) *(*to)++ = '\'';
        if  (control & C_dq) *(*to)++ = '\"';
        if  (control & C_sk) *(*to)++ = '\'';
        if  (control & C_dk) *(*to)++ = '\"';
        if  (control & C_cp) *(*to)++ = *in++;
        if  (control & C_wp) in = p;
        if  (control & C_u0) in++;
        state = control & C_state;
    }
    return(n);
};

/* Do a query and return the (0,0) value. */
char *
do_pg_query(request_rec * r, char *query, int finish) {
    PGresult *pg_result;
    char *val;
    char *result;

    result = NULL;
    pg_result = PQexec(pg_conn, query);
    if  (pg_result == NULL) {
        snprintf(pg_errstr, MAX_STRING_LEN, "auth_pg: %s", PQerrorMessage(pg_conn));
        PQfinish(pg_conn);
        pg_conn = NULL;
        return(NULL);
    }
    if  (PQresultStatus(pg_result) != PGRES_TUPLES_OK) {
        snprintf(pg_errstr, MAX_STRING_LEN, "auth_pg: %s", PQerrorMessage(pg_conn));
        PQclear(pg_result);
        PQfinish(pg_conn);
        pg_conn = NULL;
        return(NULL);
    }
    if  (PQntuples(pg_result) == 1 && !PQgetisnull(pg_result, 0, 0)) {
        val = PQgetvalue(pg_result, 0, 0);
        if  (!(result = (char *)ap_palloc(r->pool, strlen(val) + 1))) {
            snprintf(pg_errstr, MAX_STRING_LEN, "auth_pg: Could not get memory for query result");
            PQclear(pg_result);
            PQfinish(pg_conn);
            pg_conn = NULL;
            return(NULL);
        }
        strcpy(result, val);
    }
    /* ignore errors here ! */
    PQclear(pg_result);
    if  (pg_conn && finish) {
        PQfinish(pg_conn);
        pg_conn = NULL;
    }
    return(result);
}

/* Process authentication request from Apache */
int
pg_authenticate_basic_user(request_rec *r) {
    pg_auth_config_rec *sec
        = ( pg_auth_config_rec *)ap_get_module_config(r->per_dir_config, &auth_pg_module);
    char *query, *q;
    char *sent_pw, *creds;
    int res;
    size_t sums, s;

    if  ((res = ap_get_basic_auth_pw(r, (const char **)&sent_pw))) return(res);
    if  (!sec->auth_pg_function) return(DECLINED);

    pg_errstr[0] = '\0';
    sums = 25;                                                             /* include trailing '\0' */
    sums += wordscan(DOUBLE_QU, sec->auth_pg_function   , NULL, "(");
    if  (sec->auth_pg_context) sums += wordscan(SINGLE_QU, sec->auth_pg_context    , NULL, ", ");
    sums += wordscan(SINGLE_QU, r->connection->user     , NULL, ", ");
    sums += wordscan(SINGLE_QU, sent_pw                 , NULL, ", ");
    sums += wordscan(SINGLE_QU, r->connection->remote_ip, NULL, ", ");
    sums += wordscan(SINGLE_QU, r->uri                  , NULL, ", ");
    sums += wordscan(SINGLE_QU, r->args                 , NULL, ", ");
    sums += wordscan(SINGLE_QU, r->protocol             , NULL, ")");
    if  ((query = ap_palloc(r->pool, sums)) == NULL) {
        snprintf(pg_errstr, MAX_STRING_LEN, "auth_pg: No mem for query");
        return(DECLINED);
    };
    strcpy(query, "SELECT ");
    q = index(query, '\0');
    wordscan(DOUBLE_QU, sec->auth_pg_function   , &q, "(");
    if  (sec->auth_pg_context) wordscan(SINGLE_QU, sec->auth_pg_context    , &q, ", ");
    wordscan(SINGLE_QU, r->connection->user     , &q, ", ");
    wordscan(SINGLE_QU, sent_pw                 , &q, ", ");
    wordscan(SINGLE_QU, r->connection->remote_ip, &q, ", ");
    wordscan(SINGLE_QU, r->uri                  , &q, ", ");
    wordscan(SINGLE_QU, r->args                 , &q, ", ");
    wordscan(SINGLE_QU, r->protocol             , &q, ")");
    *q = '\0';
    if  (!pg_conn || PQstatus(pg_conn) != CONNECTION_OK) {
        pg_conn = PQsetdbLogin( sec->auth_pg_host
                              , sec->auth_pg_port
                              , sec->auth_pg_options
                              , NULL
                              , sec->auth_pg_database
                              , sec->auth_pg_user
                              , sec->auth_pg_pwd
                              );
        if  (!pg_conn) {
            snprintf(pg_errstr, MAX_STRING_LEN, "auth_pg: No connection to DBMS");
            return(DECLINED);
        }else if (PQstatus(pg_conn) != CONNECTION_OK) {
            snprintf(pg_errstr, MAX_STRING_LEN, "auth_pg: %s", PQerrorMessage(pg_conn));
            return(DECLINED);
    }   }
    creds = do_pg_query( r, query
                       , !r->uri || !sec->auth_pg_finish || !strcmp(r->uri, sec->auth_pg_finish)
                       );

    if  (!creds || !creds[0]) {
        if  (pg_errstr[0]) {
            res = SERVER_ERROR;
        } else {
            if  (sec->auth_pg_authoritative) {
                /* force error and access denied */
                snprintf( pg_errstr, MAX_STRING_LEN
                        , "mod_auth_pg: Password mismatch for user %s"
                        , r->connection->user
                        );
                ap_note_basic_auth_failure(r);
                res = AUTH_REQUIRED;
            } else {
                /* allow fall through to another module */
                return(DECLINED);
        }   }
        ap_log_reason(pg_errstr, r->filename, r);
        return(res);
    }
    ap_table_setn(r->subprocess_env, "CREDENTIALS", creds);
    return(OK);
}

/* Checking ID */
int
pg_check_auth(request_rec *r) {
    pg_auth_config_rec *sec
        = (pg_auth_config_rec *)ap_get_module_config(r->per_dir_config, &auth_pg_module);
    char *user = r->connection->user;
    int m = r->method_number;
    int group_result = DECLINED;
    array_header *reqs_arr = (array_header *)ap_requires(r);
    require_line *reqs = reqs_arr ? (require_line *)reqs_arr->elts : NULL;
    register int x, res;
    const char *t;
    char *w;

    pg_errstr[0] = '\0';

    /* if we cannot do it; leave it to some other guy */
    if  (!reqs_arr) {
        if  (sec->auth_pg_authoritative) {
            /* force error and access denied */
            snprintf( pg_errstr, MAX_STRING_LEN
                    , "mod_auth_pg: user %s denied, no access rules specified (PG-Authoritative)"
                    , user
                    );
            ap_log_reason(pg_errstr, r->uri, r);
            ap_note_basic_auth_failure(r);
            res = AUTH_REQUIRED;
        } else {
            return(DECLINED);
    }   }

    for (x = 0; x < reqs_arr->nelts; x++) {
        if  (!(reqs[x].method_mask & (1 << m))) continue;

        t = reqs[x].requirement;
        w = ap_getword(r->pool, &t, ' ');

        if  (!strcmp(w, "valid-user")) return(OK);

        if  (!strcmp(w, "user")) {
            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                if  (!strcmp(user, w)) return(OK);
            }
            if  (sec->auth_pg_authoritative) {
                /* force error and access denied */
                snprintf( pg_errstr, MAX_STRING_LEN
                        , "mod_auth_pg: user %s denied, no access rules specified (PG-Authoritative)"
                        , user
                        );
                ap_log_reason(pg_errstr, r->uri, r);
                ap_note_basic_auth_failure(r);
                return(AUTH_REQUIRED);
    }   }   }
    return(DECLINED);
}

_const char *
pg_set_authoritative_flag(cmd_parms *cmd, pg_auth_config_rec *sec, int arg) {
    sec->auth_pg_authoritative = arg;
    return(NULL);
}

command_rec pg_auth_cmds[] =
{   { "Auth_PG_host"
    , ap_set_string_slot
    , (void *)XtOffsetOf(pg_auth_config_rec, auth_pg_host)
    , OR_AUTHCFG
    , TAKE1
    , "the host name of the postgreSQL server."
    }
,   { "Auth_PG_database"
    , ap_set_string_slot
    , (void *)XtOffsetOf(pg_auth_config_rec, auth_pg_database)
    , OR_AUTHCFG
    , TAKE1
    , "the name of the database that contains authorization information. "
    }
,   { "Auth_PG_port"
    , ap_set_string_slot
    , (void *)XtOffsetOf(pg_auth_config_rec, auth_pg_port)
    , OR_AUTHCFG
    , TAKE1
    , "the port the postmaster is running on. "
    }
,   { "Auth_PG_options"
    , ap_set_string_slot
    , (void *)XtOffsetOf(pg_auth_config_rec, auth_pg_options)
    , OR_AUTHCFG
    , TAKE1
    , "an options string to be sent to the postgres backed process. "
    }
,   { "Auth_PG_user"
    , ap_set_string_slot
    , (void *)XtOffsetOf(pg_auth_config_rec, auth_pg_user)
    , OR_AUTHCFG
    , TAKE1
    , "user name connect as"
    }
,   { "Auth_PG_pwd"
    , ap_set_string_slot
    , (void *)XtOffsetOf(pg_auth_config_rec, auth_pg_pwd)
    , OR_AUTHCFG
    , TAKE1
    , "user pass connect as"
    }
,   { "Auth_PG_function"
    , ap_set_string_slot
    , (void *)XtOffsetOf(pg_auth_config_rec, auth_pg_function)
    , OR_AUTHCFG
    , TAKE1
    , "the name of the function to handle username/password tuples."
    }
,   { "Auth_PG_context"
    , ap_set_string_slot
    , (void *)XtOffsetOf(pg_auth_config_rec, auth_pg_context)
    , OR_AUTHCFG
    , TAKE1
    , "the name of the context."
    }
,   { "Auth_PG_finish"
    , ap_set_string_slot
    , (void *)XtOffsetOf(pg_auth_config_rec, auth_pg_finish)
    , OR_AUTHCFG
    , TAKE1
    , "request at end of page to drop PostgreSQL connect."
    }
,   { "Auth_PG_authoritative"
    , pg_set_authoritative_flag
    , NULL
    , OR_AUTHCFG
    , FLAG
    , "'on' or 'off'"
    }
,   {NULL}
};

void *
pg_auth_server_init(pool *p, server_rec *s) {
    pg_conn = NULL;
    if  (auth_pg_pool == NULL) auth_pg_pool = ap_make_sub_pool(NULL);
}

void *
create_pg_auth_dir_config(pool *p, char *d) {
    pg_auth_config_rec *new_rec;

    new_rec = ap_pcalloc(p, sizeof(pg_auth_config_rec));
    if  (new_rec == NULL) return(NULL);
    if  (auth_pg_pool == NULL) auth_pg_pool = ap_make_sub_pool(NULL);
    if  (auth_pg_pool == NULL) return(NULL);
    /* sane defaults */
    new_rec->auth_pg_host = NULL;
    new_rec->auth_pg_database = NULL;
    new_rec->auth_pg_port = NULL;
    new_rec->auth_pg_options = NULL;
    new_rec->auth_pg_user = NULL;
    new_rec->auth_pg_pwd = NULL;
    new_rec->auth_pg_function = NULL;
    new_rec->auth_pg_context = NULL;
    new_rec->auth_pg_finish = NULL;
    new_rec->auth_pg_authoritative = 1;
    return(new_rec);
}

void
pg_auth_init_handler(server_rec *s, pool *p) {
    pg_conn = NULL;
    ap_add_version_component("mod_auth_pg/" AUTH_PGSQL_VERSION);
}

module MODULE_VAR_EXPORT auth_pg_module =
{   STANDARD_MODULE_STUFF
,   pg_auth_init_handler       /* initializer                           */
,   create_pg_auth_dir_config  /* dir config creater                    */
,   NULL                       /* dir merger --- default is to override */
,   pg_auth_server_init        /* server config                         */
,   NULL                       /* merge server config                   */
,   pg_auth_cmds               /* command table                         */
,   NULL                       /* handlers                              */
,   NULL                       /* filename translation                  */
,   pg_authenticate_basic_user /* check_user_id                         */
,   pg_check_auth              /* check auth                            */
,   NULL                       /* check access                          */
,   NULL                       /* type_checker                          */
,   NULL                       /* pre-run fixups                        */
,   NULL                       /* logger                                */
,   NULL                       /* header parser                         */
,   NULL                       /* child_init                            */
,   NULL                       /* child_exit                            */
,   NULL                       /* post read-request                     */
};
